/**************************************************************************
* Otter Browser: Web browser controlled by the user, not vice-versa.
* Copyright (C) 2013 - 2014 Michal Dutkiewicz aka Emdek <michal@emdek.pl>
* Copyright (C) 2014 Jan Bajer aka bajasoft <jbajer@gmail.com>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
**************************************************************************/

#include "SearchWidget.h"
#include "PreferencesDialog.h"
#include "SearchDelegate.h"
#include "../core/SearchesManager.h"
#include "../core/SearchSuggester.h"
#include "../core/SessionsManager.h"
#include "../core/SettingsManager.h"
#include "../core/Utils.h"

#include <QtCore/QTimer>
#include <QtGui/QGuiApplication>
#include <QtGui/QMouseEvent>
#include <QtGui/QPainter>
#include <QtWidgets/QAbstractItemView>
#include <QtWidgets/QLineEdit>

namespace Otter
{

SearchWidget::SearchWidget(QWidget *parent) : QComboBox(parent),
	m_completer(new QCompleter(this)),
	m_suggester(NULL),
	m_index(0),
	m_popupUpdated(false)
{
	m_completer->setCaseSensitivity(Qt::CaseInsensitive);
	m_completer->setCompletionMode(QCompleter::PopupCompletion);
	m_completer->setCompletionRole(Qt::DisplayRole);

	installEventFilter(this);

	setEditable(true);
	setItemDelegate(new SearchDelegate(height(), this));
	setModel(SearchesManager::getSearchEnginesModel());
	setInsertPolicy(QComboBox::NoInsert);
	setCurrentSearchEngine();
	optionChanged(QLatin1String("Search/SearchEnginesSuggestions"), SettingsManager::getValue(QLatin1String("Search/SearchEnginesSuggestions")));

	lineEdit()->setCompleter(m_completer);
	lineEdit()->setStyleSheet(QLatin1String("QLineEdit {background:transparent;}"));

	connect(SearchesManager::getInstance(), SIGNAL(searchEnginesModified()), this, SLOT(storeCurrentSearchEngine()));
	connect(SearchesManager::getInstance(), SIGNAL(searchEnginesModelModified()), this, SLOT(restoreCurrentSearchEngine()));
	connect(SettingsManager::getInstance(), SIGNAL(valueChanged(QString,QVariant)), this, SLOT(optionChanged(QString,QVariant)));
	connect(this, SIGNAL(currentIndexChanged(int)), this, SLOT(currentSearchEngineChanged(int)));
	connect(this, SIGNAL(activated(int)), this, SLOT(searchEngineSelected(int)));
	connect(lineEdit(), SIGNAL(textChanged(QString)), this, SLOT(queryChanged(QString)));
	connect(lineEdit(), SIGNAL(returnPressed()), this, SLOT(sendRequest()));
	connect(m_completer, SIGNAL(activated(QString)), this, SLOT(sendRequest(QString)));
}

void SearchWidget::paintEvent(QPaintEvent *event)
{
	Q_UNUSED(event)

	QPainter painter(this);
	QStyleOptionFrame panel;
	panel.initFrom(lineEdit());
	panel.rect = rect();
	panel.palette = palette();
	panel.lineWidth = 1;

	style()->drawPrimitive(QStyle::PE_PanelLineEdit, &panel, &painter, this);

	currentData(Qt::DecorationRole).value<QIcon>().paint(&painter, m_selectButtonIconRectangle);

	QStyleOption arrow;
	arrow.initFrom(this);
	arrow.rect = m_selectButtonArrowRectangle;

	style()->drawPrimitive(QStyle::PE_IndicatorArrowDown, &arrow, &painter, this);

	Utils::getIcon(QLatin1String("edit-find")).paint(&painter, m_searchButtonRectangle);
}

void SearchWidget::resizeEvent(QResizeEvent *event)
{
	QComboBox::resizeEvent(event);

	QStyleOptionFrame panel;
	panel.initFrom(lineEdit());
	panel.rect = rect();
	panel.lineWidth = 1;

	const QRect rectangle = style()->subElementRect(QStyle::SE_LineEditContents, &panel, this);

	m_selectButtonIconRectangle = rectangle;
	m_selectButtonIconRectangle.setWidth(20);
	m_selectButtonIconRectangle = m_selectButtonIconRectangle.marginsRemoved(QMargins(2, 2, 2, 2));

	m_selectButtonArrowRectangle = rectangle;
	m_selectButtonArrowRectangle.setLeft(rectangle.left() + 20);
	m_selectButtonArrowRectangle.setWidth(12);

	m_searchButtonRectangle = rectangle;
	m_searchButtonRectangle.setLeft(rectangle.right() - 20);
	m_searchButtonRectangle = m_searchButtonRectangle.marginsRemoved(QMargins(2, 2, 2, 2));

	lineEdit()->resize((rectangle.width() - m_selectButtonIconRectangle.width() - m_selectButtonArrowRectangle.width() - m_searchButtonRectangle.width()), rectangle.height());
	lineEdit()->move(m_selectButtonArrowRectangle.topRight());

	m_lineEditRectangle = lineEdit()->geometry();

	if (event->size().height() != event->oldSize().height())
	{
		setItemDelegate(new SearchDelegate(height(), this));
	}
}

void SearchWidget::focusInEvent(QFocusEvent *event)
{
	QComboBox::focusInEvent(event);

	if (!lineEdit()->text().trimmed().isEmpty() && (event->reason() == Qt::MouseFocusReason || event->reason() == Qt::ShortcutFocusReason || event->reason() == Qt::TabFocusReason || event->reason() == Qt::BacktabFocusReason) && SettingsManager::getValue(QLatin1String("AddressField/SelectAllOnFocus")).toBool())
	{
		QTimer::singleShot(0, lineEdit(), SLOT(selectAll()));
	}
	else if (event->reason() != Qt::PopupFocusReason)
	{
		lineEdit()->deselect();
	}
}

void SearchWidget::mousePressEvent(QMouseEvent *event)
{
	QWidget::mousePressEvent(event);
}

void SearchWidget::mouseReleaseEvent(QMouseEvent *event)
{
	if (m_selectButtonIconRectangle.contains(event->pos()) || m_selectButtonArrowRectangle.contains(event->pos()))
	{
		showPopup();
	}
	else if (m_searchButtonRectangle.contains(event->pos()))
	{
		sendRequest();
	}
	else
	{
		QWidget::mouseReleaseEvent(event);
	}
}

void SearchWidget::wheelEvent(QWheelEvent *event)
{
	disconnect(lineEdit(), SIGNAL(textChanged(QString)), this, SLOT(queryChanged(QString)));

	QComboBox::wheelEvent(event);

	lineEdit()->setText(m_query);

	connect(lineEdit(), SIGNAL(textChanged(QString)), this, SLOT(queryChanged(QString)));
}

void SearchWidget::hidePopup()
{
	if (!m_query.isEmpty())
	{
		m_popupUpdated = true;
	}

	QComboBox::hidePopup();
}

void SearchWidget::optionChanged(const QString &option, const QVariant &value)
{
	if (option == QLatin1String("Search/SearchEnginesSuggestions"))
	{
		if (value.toBool() && !m_suggester)
		{
			m_suggester = new SearchSuggester(getCurrentSearchEngine(), this);

			m_completer->setModel(m_suggester->getModel());

			connect(lineEdit(), SIGNAL(textEdited(QString)), m_suggester, SLOT(setQuery(QString)));
		}
		else if (!value.toBool() && m_suggester)
		{
			m_suggester->deleteLater();
			m_suggester = NULL;

			m_completer->setModel(NULL);
		}
	}
}

void SearchWidget::currentSearchEngineChanged(int index)
{
	if (!setPlaceholderText(index))
	{
		return;
	}

	if (!m_query.isEmpty())
	{
		sendRequest();
	}
}

void SearchWidget::searchEngineSelected(int index)
{
	if (!m_storedSearchEngine.isEmpty())
	{
		return;
	}

	if (itemData(index, Qt::AccessibleDescriptionRole).toString().isEmpty())
	{
		m_index = index;

		if (m_suggester)
		{
			m_suggester->setEngine(getCurrentSearchEngine());
			m_suggester->setQuery(QString());
		}

		SessionsManager::markSessionModified();
	}
	else
	{
		const QString query = m_query;

		setCurrentIndex(m_index);

		if (query != itemText(index))
		{
			lineEdit()->setText(query);
		}

		if (itemData(index, Qt::AccessibleDescriptionRole).toString() == QLatin1String("configure"))
		{
			PreferencesDialog dialog(QLatin1String("search"), this);
			dialog.exec();
		}
	}
}

void SearchWidget::queryChanged(const QString &query)
{
	if (m_popupUpdated)
	{
		m_popupUpdated = false;
	}
	else
	{
		m_query = query;
	}
}

void SearchWidget::sendRequest(const QString &query)
{
	if (!query.isEmpty())
	{
		m_query = query;
	}

	if (!m_query.isEmpty())
	{
		emit requestedSearch(m_query, currentData(Qt::UserRole + 1).toString(), (QGuiApplication::keyboardModifiers().testFlag(Qt::ShiftModifier) ?  NewTabOpen : DefaultOpen));
	}
}

void SearchWidget::storeCurrentSearchEngine()
{
	m_storedSearchEngine = getCurrentSearchEngine();

	hidePopup();

	disconnect(this, SIGNAL(currentIndexChanged(int)), this, SLOT(currentSearchEngineChanged(int)));
}

void SearchWidget::restoreCurrentSearchEngine()
{
	if (!m_storedSearchEngine.isEmpty())
	{
		setCurrentSearchEngine(m_storedSearchEngine);

		m_storedSearchEngine = QString();
	}

	connect(this, SIGNAL(currentIndexChanged(int)), this, SLOT(currentSearchEngineChanged(int)));
}

void SearchWidget::setCurrentSearchEngine(const QString &engine)
{
	const QStringList engines = SearchesManager::getSearchEngines();

	if (engines.isEmpty())
	{
		hidePopup();
		setEnabled(false);

		lineEdit()->setPlaceholderText(QString());

		return;
	}

	const int index = qMax(0, engines.indexOf(engine.isEmpty() ? SettingsManager::getValue(QLatin1String("Search/DefaultSearchEngine")).toString() : engine));

	setCurrentIndex(index);
	setPlaceholderText(index);
	searchEngineSelected(index);
	setEnabled(true);

	if (m_suggester)
	{
		m_suggester->setEngine(getCurrentSearchEngine());
	}
}

QString SearchWidget::getCurrentSearchEngine() const
{
	return currentData(Qt::UserRole + 1).toString();
}

bool SearchWidget::setPlaceholderText(int index)
{
	lineEdit()->setGeometry(m_lineEditRectangle);

	if (itemData(index, Qt::AccessibleDescriptionRole).toString().isEmpty())
	{
		lineEdit()->setPlaceholderText(tr("Search Using %1").arg(itemData(index, Qt::UserRole).toString()));
	}
	else
	{
		lineEdit()->setPlaceholderText(QString());

		return false;
	}

	lineEdit()->setText(m_query);

	return true;
}

bool SearchWidget::eventFilter(QObject *object, QEvent *event)
{
	if (object == this && event->type() == QEvent::KeyPress)
	{
		QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);

		if ((keyEvent->key() == Qt::Key_Enter || keyEvent->key() == Qt::Key_Return) && keyEvent->modifiers() & Qt::ShiftModifier)
		{
			const QString input = lineEdit()->text().trimmed();

			if (!input.isEmpty())
			{
				m_query = input;
			}

			if (!m_query.isEmpty())
			{
				emit requestedSearch(m_query, currentData(Qt::UserRole + 1).toString(), NewTabOpen);
			}

			return true;
		}
	}

	return QComboBox::eventFilter(object, event);
}

}
